"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.efetch.efetch_request
  :synopsis: Exports class EfetchRequest implementing individual requests for
    entrezpy queries to NCBI Efetch Eutils

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import entrezpy.base.request


class EfetchRequest(entrezpy.base.request.EutilsRequest):
  """The EfetchRequest class implements a single request as part of an Efetch
  query. It stores and prepares the parameters for a single request.
  :meth:`entrezpy.efetch.efetch_query.Efetch.inquire` calculates start and size
  for a single request.

  :param parameter: request parameter
  :param type: :class:`entrezpy.efetch.efetch_parameter.EfetchParameter`
  :param int start: number of first UID to fetch
  :param int size: requets size
  """

  def __init__(self, eutil, parameter, start, size):
    super().__init__(eutil, parameter.db)
    self.start = start
    self.retmax = size
    self.uids = parameter.uids[start:start+self.retmax]
    self.webenv = parameter.webenv
    self.querykey = parameter.querykey
    self.rettype = parameter.rettype
    self.retmode = parameter.retmode
    self.strand = parameter.strand
    self.seqstart = parameter.seqstart
    self.seqstop = parameter.seqstop
    self.complexity = parameter.complexity

  def get_post_parameter(self):
    qry = self.prepare_base_qry()
    if self.retmode != None:
      qry.update({'retmode' : self.retmode})
    if self.rettype != None:
      qry.update({'rettype' : self.rettype})
    if self.strand != None:
      qry.update({'strand' : self.strand})
    if self.seqstart != None:
      qry.update({'seq_start' : self.seqstart})
    if self.seqstop != None:
      qry.update({'seq_stop' : self.seqstop})
    if self.complexity != None:
      qry.update({'complexity' : self.complexity})
    if self.webenv and self.querykey:
      qry.update({'WebEnv' : self.webenv, 'query_key' : self.querykey,
                  'retstart' : self.start, 'retmax' : self.retmax})
    else:
      qry.update({'id' : ','.join(str(x) for x in self.uids)})
    return qry

  def dump(self):
    """Dumps instance attributes"""
    return {'db' : self.db,
            'uids' : self.uids,
            'num_uids' : len(self.uids),
            'webenv' : self.webenv,
            'querykey' : self.querykey,
            'rettype' : self.rettype,
            'retmode' :self.retmode,
            'retmax' : self.retmax,
            'retstart' : self.start,
            'strand' : self.strand,
            'seqstart' : self.seqstart,
            'seqstop' : self.seqstop,
            'complexity' : self.complexity}

  def get_observation(self):
    """Overwrite :meth:`entrezpy.base.request.EutilsRequest.get_observation`
    for Efetch requests"""
    cols = [self.query_id, self.id, self.start, self.retmax, self.status, self.duration]
    if self.request_error != None:
      cols.append(self.request_error)
    return '\t'.join(str(x) for x in cols)



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.base.monitor
  :synopsis: Exports class QueryMonitor implementing monitoring of entrezpy
    queries and requsts

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import sys
import threading
import logging
import time


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())


class QueryMonitor:
  """The QueryMonitor class implements the monitoring of entrezpy requests in
  entrezpy queries. It controls individual Observers which are linked to one
  entrezpy query.
  """

  class Observer(threading.Thread):
    """The Observer class implements the observation of one entrezpy query. It
    uses Python's multithreading deamon when using multithreading."""
    def __init__(self):
      super().__init__(daemon=True)
      self.expected_requests = 0
      self.processed_requests = 0
      self.doObserve = True
      self.requests = []
      self.duration = None

    def recall(self):
      """Stops an observation for a query"""
      self.doObserve = False
      print('\n', file=sys.stderr)
      for i in self.requests:
        print(i.get_observation(), file=sys.stderr)
      self.join()

    def dispatch(self, parameter):
      """Starts observation

      :param parameter: query parameter
      :type  parameter: 'class':`entrezpy.base.EutilsParameter`
      """
      self.expected_requests = parameter.expected_requests
      if not self.is_alive():
        self.start()

    def observe(self, request):
      """Adds one query request for observation"""
      self.requests.append(request)

    def run(self):
      """Observes requests from an entrezpy query"""
      while self.doObserve:
        for i in self.requests:
          print("{0}/{1}\t{2}".format(self.processed_requests,
                                      self.expected_requests,
                                      i.get_observation()), end='\r', file=sys.stderr)
        time.sleep(1)

  def __init__(self):
    """:ivar dict observers: observer storage for queries"""
    self.observers = {}
    #self.locks = {}

  def register_query(self, query):
    """Adds a query for observation

    :param query: entrezpy query
    :type  query: :class:`entrezpy.base.query.EutilsQuery`
    """
    #self.locks[query.id] = threading.Lock()
    self.observers[query.id] = self.Observer()

  def get_observer(self, query_id):
    """Returns an observer for a specific query.

    :param str query_id: entrezpy query id
    :rtype: :class:`base.mnonitor.QueryMonitor.Observer`
    """
    return self.observers.get(query_id, None)

  def dispatch_observer(self, query, parameter):
    """Start the observer for an entrezpy query

    :param query: entrezpy query
    :type  query: :class:`entrezpy.base.query.EutilsQuery`
    :param parameter: query parameter
    :type  parameter: 'class':`entrezpy.base.EutilsParameter`
    """
    observer = self.observers.get(query.id, None)
    if observer:
      observer.dispatch(parameter)

  def recall_observer(self, query):
    """Stops the observer for an entrezpy query

    :param query: entrezpy query
    :type  query: :class:`entrezpy.base.query.EutilsQuery`
    """
    observer = self.observers.get(query.id, None)
    if observer:
      observer.recall()

  def update_observer(self, query, parameter):
    """ Function updating the settings for a thread. Honestly, I have no idea if
    the lock is really required. It works without locks, just updating the
    parameter.

    :param query: entrezpy query
    :type  query: :class:`entrezpy.base.query.EutilsQuery`
    :param parameter: query parameter
    :type  parameter: 'class':`entrezpy.base.EutilsParameter`
    """
    self.observers[query.id].expected_requests = parameter.expected_requests



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.wally
   :synopsis: Exports class Wally implementing entrezpy pipelines to query
      NCBI E-Utilities

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""

import sys
import json
import uuid
import base64
import queue
import logging

import entrezpy.esearch.esearcher
import entrezpy.esearch.esearch_analyzer
import entrezpy.elink.elinker
import entrezpy.elink.elink_analyzer
import entrezpy.epost.eposter
import entrezpy.epost.epost_analyzer
import entrezpy.efetch.efetcher
import entrezpy.esummary.esummarizer
import entrezpy.esummary.esummary_analyzer

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


class Wally:
  """Wally simplifies to create pipelines and queries for entrezpy. Wally stores
  results from previous requests, allowing to concatenate queries and
  retrieve obtained results later if required to reduce the need to redownload
  data. Wally can use multiple threads to speed up data download, but some
  external libraries can break, e.g. SQLite3.

  Pipelines
  ---------
  Queries instances in pipelines of :class:`Wally.Pipeline` are stored in the
  dictionary :attr:`Wally.queries` with the query id as key and are accessible
  by all Wally instances. A single :class:`Wally.Pipeline` stores only the
  query id for this instance

  :param str email: user email
  :param str apikey: NCBI apikey
  :param str apikey_var: enviroment variable storing NCBI apikey
  :param int threads: set threads for multithreading
  """

  queries = {}
  """Query storage"""

  analyzers = {}
  """Analyzed query storage"""

  class Query:
    """ Entrezpy query for a Wally pipeline. Wally assembles pipelines using
    several Query() instances. If a dependency is given, it uses those
    parameters as basis using `:meth:`.resolve_dependency`.

    :param str function: Eutils function
    :param dict parameter: function parameters
    :param str dependency: query id from earlier query
    :param analyzer: analyzer instance for this query
    :type analyzer: :class:`entrezpy.base.analyzer.EutilsAnalyzer`
    """

    def __init__(self, function, parameter, dependency=None, analyzer=None):
      if not parameter and not dependency:
        sys.exit(logger.error(json.dumps({__name__:{'Error': 'Missing required parameters' \
                                                    'parameter and/or `dependency`.',
                                                    'action' : 'abort'}})))
      if not parameter:
        parameter = {}
      self.id = base64.urlsafe_b64encode(uuid.uuid4().bytes).decode()
      self.function = function
      self.parameter = parameter
      self.dependency = dependency
      self.analyzer = analyzer

    def resolve_dependency(self):
      """Resolves dependencies to obtain paremeters from earlier query.
      Parameters passed to this instance will overwrite dependency parameters
      """
      if self.dependency:
        parameter = Wally.analyzers[self.dependency].result.get_link_parameter()
        if self.function == 'elink':
          parameter['dbfrom'] = parameter['db']
        parameter.update(self.parameter)
        self.parameter = parameter

  class Pipeline:
    """The Pipeline class implements a query pipeline with several consecutive
    queries. New pipelines are obtained through :class:`.Wally`. Query instances
    are stored in :attr:`Wally.queries` and the corresponding query id's in
    :attr:`.queries`. Every added query returns its id which can be used to
    retrieve it.
    """

    def __init__(self):
      """:ivar  queries: queries for this Pipeline instance
         :type  queries: :class:`queue.Queue()`
      """
      self.queries = queue.Queue()

    def add_search(self, parameter=None, dependency=None, analyzer=None):
      """Adds Esearch query

      :param dict parameter: Esearch E-Eutility parameters
      :param str dependency: query id from earlier query
      :param analyzer: analyzer for this query
      :type  analyzer: :class:`entrezpy.base.analyzer.EutilsAnalyzer`
      :return: Wally query
      :rtype: :class:`WallyQuery`
      """
      return self.add_query(Wally.Query('esearch', parameter, dependency, analyzer))

    def add_link(self, parameter=None, dependency=None, analyzer=None):
      """Adds Elink query. Signature as :meth:`Wally.Pipeline.add_search`"""
      return self.add_query(Wally.Query('elink', parameter, dependency, analyzer))

    def add_post(self, parameter=None, dependency=None, analyzer=None):
      """Adds Epost query. Signature as :meth:`Wally.Pipeline.add_search`"""
      return self.add_query(Wally.Query('epost', parameter, dependency, analyzer))

    def add_summary(self, parameter=None, dependency=None, analyzer=None):
      """Adds Esummary query. Signature as :meth:`Wally.Pipeline.add_search`"""
      return self.add_query(Wally.Query('esummary', parameter, dependency, analyzer))

    def add_fetch(self, parameter=None, dependency=None, analyzer=None):
      """Adds Efetch query. Same signature as :meth:`Wally.Pipeline.add_search`
      but analyzer is required as this step obtains highly variable results.
      """
      #if not analyzer:
        #logger.error(json.dumps({__name__ : {'Error' : {'Missing required parameter' : 'analyzer',
                                                        #'action' : 'abort'}}}))
      return self.add_query(Wally.Query('efetch', parameter, dependency, analyzer))

    def add_query(self, query):
      """Adds query to own pipeline and storage

      :param query: Wally query
      :type  query: :class:`Wally.Query`
      :return: query id of added query
      :rtype: str
      """
      self.queries.put(query.id)
      Wally.queries[query.id] = query
      return query.id

  def __init__(self, email, apikey=None, apikey_envar=None, threads=None):
    self.tool = 'wally'
    self.email = email
    self.apikey = apikey
    self.api_envar = apikey_envar
    self.threads = threads

  def run(self, pipeline):
    """Runs one query in pipeline and checks for errors. If errors are
    encounterd the pipeline aborts.

    :param pipeline: Wally pipeline
    :type  pipeline: :class:`Wally.Pipeline`
    """
    while not pipeline.queries.empty():
      q = Wally.queries[pipeline.queries.get()]
      q.resolve_dependency()
      logger.info(json.dumps({__name__ : {'Inquiring' : {'query_id' : q.id,
                                                         'function' : q.function}}}))
      if q.function == 'esearch':
        Wally.analyzers[q.id] = self.search(q)
      if q.function == 'elink':
        Wally.analyzers[q.id] = self.link(q)
      if q.function == 'efetch':
        Wally.analyzers[q.id] = self.fetch(q)
      if q.function == 'epost':
        Wally.analyzers[q.id] = self.post(q)
      if q.function == 'esummary':
        Wally.analyzers[q.id] = self.summarize(q)
      self.check_query(q)
      if Wally.analyzers[q.id].isEmpty():
        logger.info(json.dumps({__name__ : {'empty response': {'query_id' : q.id,
                                                               'action' : 'skip'}}}))
        return Wally.analyzers[q.id]
    return Wally.analyzers[q.id]

  def check_query(self, query):
    """Check for successful query.

    :param query: Wally query
    :type  query: :class:`Wally.Query`
    """
    if not Wally.analyzers[query.id]:
      sys.exit(logger.error(json.dumps({__name__ : {'Request error': {'query_id' : query.id,
                                                                      'action' : 'abort'}}})))
    if not Wally.analyzers[query.id].isSuccess():
      sys.exit(logger.info(json.dumps({__name__ : {'response error': {'query_id' : query.id,
                                                                      'action' : 'abort'}}})))

  def get_result(self, query_id):
    """"Returns stored result from previous run.

    :param str query_id: query id
    :return: Result from this query
    :rtype: :class:`entrezpy.base.result.EutilsResult`
    """
    analyzer = self.analyzers.get(query_id)
    if not analyzer:
      return None
    return analyzer.get_result()

  def new_pipeline(self):
    """Retrurns new Wally pipeline.

    :return: Wally pipeline
    :rtype: :class:`Wally.Pipeline`
    """
    return self.Pipeline()

  def search(self, query, analyzer=entrezpy.esearch.esearch_analyzer.EsearchAnalyzer):
    """Configures and runs an Esearch query. Analyzer are class references and
    instantiated here.

    :param query: Wally Query
    :type  query: :class:`Wally.Query`
    :param analyzer: reference to analyzer class
    :return: analyzer
    :rtype: :class:`entrezpy.esearch.esearch_analyzer.EsearchAnalyzer`
    """
    analyzer = query.analyzer if query.analyzer else analyzer()
    return entrezpy.esearch.esearcher.Esearcher(self.tool,
                                                self.email,
                                                self.apikey,
                                                threads=self.threads,
                                                qid=query.id).inquire(query.parameter, analyzer)

  def summarize(self, query, analyzer=entrezpy.esummary.esummary_analyzer.EsummaryAnalzyer):
    """Configures and runs an Esummary query. Analyzer are class references and
    instantiated here.

    :param query: Wally Query
    :type  query: :class:`Wally.Query`
    :param analyzer: reference to analyzer class
    :return: analyzer
    :rtype: :class:`entrezpy.esummary.esummary_analyzer.EsummaryAnalyzer`
    """
    analyzer = query.analyzer if query.analyzer else analyzer()
    return entrezpy.esummary.esummarizer.Esummarizer(self.tool,
                                                     self.email,
                                                     self.apikey,
                                                     threads=self.threads,
                                                     qid=query.id).inquire(query.parameter, analyzer)

  def link(self, query, analyzer=entrezpy.elink.elink_analyzer.ElinkAnalyzer):
    """Configures and runs an Elink query. Analyzer are class references and
    instantiated here.

    :param query: Wally Query
    :type  query: :class:`Wally.Query`
    :param analyzer: reference to analyzer class
    :return: analyzer
    :rtype: :class:`entrezpy.elink.elink_analyzer.ElinkAnalyzer`
    """
    analyzer = query.analyzer if query.analyzer else analyzer()
    return entrezpy.elink.elinker.Elinker(self.tool,
                                          self.email,
                                          self.apikey,
                                          threads=self.threads,
                                          qid=query.id).inquire(query.parameter, analyzer)

  def post(self, query, analyzer=entrezpy.epost.epost_analyzer.EpostAnalyzer):
    """Configures and runs an Epost query. Analyzer are class references and
    instantiated here.

    :param query: Wally Query
    :type  query: :class:`Wally.Query`
    :param analyzer: reference to analyzer class
    :return: analyzer
    :rtype: :class:`entrezpy.epost.epost_analyzer.EpostAnalyzer`
    """
    analyzer = query.analyzer if query.analyzer else analyzer()
    return entrezpy.epost.eposter.Eposter(self.tool,
                                          self.email,
                                          self.apikey,
                                          threads=self.threads,
                                          qid=query.id).inquire(query.parameter, analyzer)

  def fetch(self, query, analyzer=entrezpy.efetch.efetch_analyzer.EfetchAnalyzer):
    """uns an Efetch query. The Analyzer needs to be added to the quuery

    :param query: Wally Query
    :type  query: :class:`Wally.Query`
    :param analyzer: reference to analyzer class
    :return: analyzer
    :return: analyzer
    :rtype: :class:`entrezpy.efetch.efetch_analyzer.EfetchAnalyzer`
    """
    analyzer = query.analyzer if query.analyzer else analyzer()
    return entrezpy.efetch.efetcher.Efetcher(self.tool,
                                             self.email,
                                             self.apikey,
                                             threads=self.threads,
                                             qid=query.id).inquire(query.parameter, analyzer)



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.elink.linkset.unit.linkout_all_attributes
   :synopsis: Exports :class:`LinkOutAllAttributes` implementing Elink results
    for `llinkslib` command reporting URLs and attributes for all links outside
    Entrez.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import entrezpy.elink.linkset.unit.linksetunit


class LinkOutAllAttribute(entrezpy.elink.linkset.unit.linksetunit.LinksetUnit):
  """ The `LinkOutAllAttribute` class represents a result from the `llinkslib`
  Elink command. Results show the URL and its attributes for all links outside
  Entrez.

  :param dict unit: Entrez link data
  """

  @classmethod
  def new(cls, unit):
    """:rtype: :class:`linkout_all_attributes.LinkOutAllAttributes`"""
    return cls(unit)

  class Provider:
    """The Provider class stores all attrbiutes for an URL link outside Entrez

    :param dict provider_obj: attributes for linked URL
    """
    def __init__(self, provider_obj):
      self.id = provider_obj.pop('id', None)
      self.name = provider_obj.pop('name', None)
      self.nameabbr = provider_obj.pop('nameabbr', None)
      self.url = provider_obj.pop('url', None)
      self.iconurl = provider_obj.pop('iconurl', None)

    def dump(self):
      ":rtype: dict"
      return {'id' : self.id, 'name' : self.name, 'nameabbr' : self.nameabbr,
              'url' : self.url, 'iconurl' : self.iconurl}

  def __init__(self, unit):
    super().__init__(None, unit.pop('linkname', None))
    self.iconurl = unit.pop('iconurl', None)
    self.subjecttype = unit.pop('subjecttype', None)
    self.category = unit.pop('category', None)
    self.attributes = unit.pop('attributes', [])
    self.provider = self.Provider(unit.pop('provider', {}))

  def dump(self):
    ":rtype: dict"
    return dict({'iconurl' : self.iconurl, 'provider' : self.provider.dump(),
                 'subjecttype' : self.subjecttype, 'category' : self.category,
                 'attributes' : self.attributes, 'linkname' : self.linkname},
                **self.basic_dump())



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.requester.requester
  :synopsis: Exports class Requester handling HTTP requests for entrezpy.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import sys
import random
import json
import time
import socket
import logging
import urllib.parse
import urllib.request
import urllib.error

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())


class Requester:
  """Requester implements the sendong of HTTP POST requests and the receiving
  of the result. It checks for request connection errors and performs retries
  when possible. If the maximum number of retries is reached, the request is
  conisdered failed.  In case of connections errors, abort if the error is not
  due to timeout. The initial timeout is increased insteps until the maximum
  timeout has been reached.

  :param float wait: wait time in seconds between requests
  :param int max_retries: number of rertries before giving up.
  :param int init_timeout: number of seconds before the initial request is consid
                           considered a timeout error
  :param int timeout_max: maximum requet timeout before giving up
  :param int timeout_steps: increase value for timeout errors
  """
  def __init__(self, wait, max_retries=9, init_timeout=10, timeout_max=60, timeout_step=5):
    self.wait = wait
    self.max_retries = max_retries
    self.init_timeout = init_timeout
    self.timeout_max = timeout_max
    self.timeout_step = timeout_step
    logger.info(json.dumps({__name__ : {'settings' : {'wait[s]' : self.wait,
                                                      'timeout-min[s]' : self.init_timeout,
                                                      'timeout-max[s]' : self.timeout_max,
                                                      'timeout-increase[s]' : self.timeout_step,
                                                      'max-retries' : self.max_retries}}}))
  def request(self, req):
    """Request the request

    :param req: entrezpy request
    :type  req: :class:`entrezpy.base.request.EutilsRequest`
    """
    retries = 0
    success = False
    req_timeout = self.init_timeout
    while not success:
      wait = self.wait
      try:
        data = urllib.parse.urlencode(req.get_post_parameter(), doseq=req.doseq).encode('utf-8')
        req.qry_url = data.decode()
        logger.debug(json.dumps({__name__ : {'Request' : {'try' : retries,
                                                          'req-id' : req.id,
                                                          'req-query' : req.query_id,
                                                          'req-url' : req.url,
                                                          'qry-url' : req.qry_url}}}))
        req.set_status_success()
        return urllib.request.urlopen(urllib.request.Request(req.url, data=data),
                                      timeout=req_timeout)
      except urllib.error.HTTPError as http_err:
        log_msg = {'HTTP-error': {'code' : http_err.code, 'reason' : http_err.reason}}
        if http_err.code == 400: # Bad request form, stop right now
          log_msg['HTTP-error'].update({'action' : 'abort'})
          logger.debug(json.dumps({__name__ : {'Request-error' : log_msg}}))
          sys.exit()
        log_msg['HTTP-error'].update({'action' : 'retry'})
        logger.debug(json.dumps({__name__ : {'HTTP-error': log_msg}}))
        retries += 1
        wait = random.randint(1, 3)
      except urllib.error.URLError as url_err:
        req.set_request_error(url_err.reason)
        print(url_err.reason)
        logger.debug(json.dumps({__name__ : {'Request-error' : {'URL-error' : url_err.reason,
                                                                'action' : 'retry'}}}))
        retries += 1
        wait = random.randint(1, 3)
      except socket.timeout:
        req_timeout += self.timeout_step
        logger.info(json.dumps({__name__ : {'Timeout' :{'action' : 'retry',
                                                        'wait' : wait,
                                                        'timeout' : req_timeout,
                                                        'increase' :self.timeout_step}}}))
        retries += 1
        if req_timeout > self.timeout_max:
          logger.info(json.dumps({__name__: {'MaxTimeout' : {'action' : 'giving up on request',
                                                             'timeout': req_timeout}}}))
          req.set_request_error("MaxTimeout")
          return None
      else:
        if retries > self.max_retries:
          logger.info(json.dumps({__name__ : {'MaxRetry' : {'action' : 'giving up on this request',
                                                            'retries' : retries,
                                                            'max' : self.max_retries}}}))
          req.set_request_error("MaxRetry")
          return None
        success = True
      time.sleep(wait)



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.esummary.esummary_request
  :synopsis: Exports class EsummaryRequest implementing individual requests for
    entrezpy queries to Esummary NCBI E-Utility

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import entrezpy.base.request


class EsummaryRequest(entrezpy.base.request.EutilsRequest):
  """The SummaryRequest class implements a single request as part of an Esummary
  query. It stores and prepares the parameters for a single request.
  :meth:`entrezpy.esummary.esummary_query.Esummaryizer.inquire` calculates start
  and size for a single request.

  :param parameter: request parameter
  :param type: :class:`entrezpy.esummary.esummary_parameter.EsummaryParameter`
  :param int start: number of first UID to fetch
  :param int size: requets size
  """

  def __init__(self, eutil, parameter, start, size):
    super().__init__(eutil, parameter.db)
    self.retstart = start
    self.retmax = size
    self.retmode = parameter.retmode
    self.rettype = parameter.rettype
    self.uids = parameter.uids[start:start+size]
    self.webenv = parameter.webenv
    self.querykey = parameter.querykey

  def get_post_parameter(self):
    qry = self.prepare_base_qry(extend={'retmode':self.retmode})
    if self.webenv and self.querykey:
      qry.update({'WebEnv' : self.webenv, 'query_key':self.querykey,
                  'retstart' : self.retstart, 'retmax' : self.retmax})
    else:
      qry.update({'id' : ','.join(str(x) for x in self.uids)})
    return qry

  def dump(self):
    """:rtype: dict"""
    return self.dump_internals({'retstart' : self.retstart,
                                'retmax' : self.retmax,
                                'retmode' : self.retmode,
                                'rettype' : self.rettype,
                                'WebEnv' : self.webenv,
                                'query_key' : self.querykey,
                                'uids' : len(self.uids)})



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.esummary.esummarizer
   :synopsis: Exports class Esummarizer implementing entrezpy Esummary queries
    to NCBI E-Utility Esummary

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json
import logging

import entrezpy.base.query
import entrezpy.esummary.esummary_request
import entrezpy.esummary.esummary_analyzer
import entrezpy.esummary.esummary_parameter


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())


class Esummarizer(entrezpy.base.query.EutilsQuery):
  """Esummary implements Esummary queries to E-Utilities [0]. It fetches
  summaries for given UIDs or Webenv and query_key references.
  Esummary can consist of several queries, depending on the format requested.
  # [0]: https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ESummary
  """

  def __init__(self, tool, email, apikey=None, apikey_var=None, threads=None, qid=None):
    super().__init__('esummary.fcgi', tool, email, apikey, apikey_var, threads, qid)

  def inquire(self, parameter, analyzer=entrezpy.esummary.esummary_analyzer.EsummaryAnalzyer()):
    """Implements :meth:`entrezpy.base.quet.EutilsQuery.unquire`.

    :param dict parameter: Esummary parameter
    :param analyzer: Esummary analyzer
    :type  analyzer: :class:`entrezpy.esummary.esummary_analzyer`
    :return: analyzer instance or None if request errors have been encountered
    :rtype: :class:`entrezpy.esummary.esummary_analyzer.EsummaryAnalyzer` or
      None
    """
    param = entrezpy.esummary.esummary_parameter.EsummaryParameter(parameter)
    logger.debug(json.dumps({__name__ : {'Parameter' : param.dump()}}))
    req_size = param.reqsize
    self.monitor_start(param)
    for i in range(param.expected_requests):
      if i * req_size + req_size > param.retmax:
        req_size = param.retmax % param.reqsize
      self.add_request(entrezpy.esummary.esummary_request.EsummaryRequest(self.eutil,
                                                                          param,
                                                                          (i*param.reqsize),
                                                                          req_size), analyzer)
    self.request_pool.drain()
    self.monitor_stop()
    if self.check_requests() == 0:
      return analyzer
    return None

  def check_requests(self):
    """Test for request errors

      :return: 1 if request errors else 0
      :rtype: int
    """
    if not self.hasFailedRequests():
      logger.info(json.dumps({__name__ : {'Query status' : {self.id : 'OK'}}}))
      return 0
    logger.info(json.dumps({__name__ : {'Query status' : {self.id : 'failed'}}}))
    logger.debug(json.dumps({__name__ : {'Query status' :
                                         {self.id : 'failed',
                                          'request-dumps' : [x.dump_internals()
                                                             for x in self.failed_requests]}}}))
    return 1



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: elink_request
  :synopsis:
    Exports class ElinkRequest implementing individual requests as part of a
            entrezpy query.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""

import entrezpy.base.request

class ElinkRequest(entrezpy.base.request.EutilsRequest):
  """ The ElinkRequest class implements a single request as part of a Elinker
  query. It stores and prepares the parameters for a single request.
  See :class:`entrezpy.elink.elink_parameter.ElinkParameter` for parameter
  description.

  :param parameter: request parameter
  :param type: :class:`entrezpy.elink.elink_parameter.ElinkParameter`
  """
  def __init__(self, eutil, parameter):
    super().__init__(eutil, parameter.db)
    self.dbfrom = parameter.dbfrom
    self.cmd = parameter.cmd
    self.querykey = parameter.querykey
    self.webenv = parameter.webenv
    self.uids = parameter.uids
    self.retmode = parameter.retmode
    self.linkname = parameter.linkname
    self.term = parameter.term
    self.holding = parameter.holding
    self.datetype = parameter.datetype
    self.reldate = parameter.reldate
    self.mindate = parameter.mindate
    self.maxdate = parameter.maxdate
    self.doseq = parameter.doseq

  def get_post_parameter(self):
    """Implements :meth:`entrezpy.base.request.EutilsRequest.get_post_parameter`.

    - If `WebEnv` and `query_key` are given the history server will be used.
    - If UIDs are given create an id parameter for each UID, i.e. id=123&id=456
      (see :attr:`entrezpy.elink.elink.elink_parameter.ElinkParameter.doseq`)
    - Setting :attr:`entrezpy.elink.elink.elink_parameter.ElinkParameter.doseq`
      to `False` concatenats UIDs with commas, i.e. id=123,456

    `linkname`: For `neighbor` or `neighbor` commands without a given linkname
     one generated. See documentation for more details.
    """
    qry = self.prepare_base_qry(extend={'cmd':self.cmd, 'retmode':self.retmode,
                                        'dbfrom':self.dbfrom, 'db':self.db})
    if self.webenv and self.querykey:
      qry.update({'WebEnv' : self.webenv, 'query_key' : self.querykey})
    else:
      if self.doseq:
        qry.update({'id' : self.uids})
      else:
        qry.update({'id' : ','.join(str(x) for x in self.uids)})
    if self.cmd == 'neighbor' or self.cmd == 'neighbor_history' or self.cmd == 'neighbor_score':
      if not self.linkname:
        self.linkname = '_'.join([self.dbfrom, self.db])
      qry.update({'linkname' : self.linkname})
    if self.term:
      qry.update({'term' : self.term})
    if self.holding:
      qry.update({'holding' : self.holding})
    if self.datetype:
      qry.update({'datetype' : self.datetype})
    if self.reldate:
      qry.update({'reldate' : self.reldate})
    if self.mindate:
      qry.update({'mindate' : self.mindate})
    if self.maxdate:
      qry.update({'maxdate' : self.maxdate})
    if  not self.db:
      qry.pop('db')
    return qry

  def dump(self):
    """Dumps instance attributes

    :return: instance attributes
    :rtype: dict
    """
    return self.dump_internals({'retmode' : self.retmode,
                                'WebEnv' : self.webenv,
                                'cmd' : self.cmd,
                                'holding' : self.holding,
                                'term' : self.term,
                                'query_key' : self.querykey,
                                'dbfrom' : self.dbfrom,
                                'datetype' : self.datetype,
                                'reldate' : self.reldate,
                                'mindate' : self.mindate,
                                'maxdate' : self.maxdate})



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.efetch.efetcher
  :synopsis: Exports class Efetcher implementing entrezpy Efetch queries to
    NCBI EUtils Efetch

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json
import logging

import entrezpy.base.query
import entrezpy.efetch.efetch_parameter
import entrezpy.efetch.efetch_request
import entrezpy.efetch.efetch_analyzer


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())


class Efetcher(entrezpy.base.query.EutilsQuery):
  """Efetcher implements Efetch E-Utilities queries [0]. It implements
  :meth:`entrezpy.base.query.EutilsQuery.inquire` to fetch data from NCBI
  Entrez servers.
  [0]: https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EFetch
  [1]: https://www.ncbi.nlm.nih.gov/books/NBK25497/table/
  chapter2.T._entrez_unique_identifiers_ui/?report=objectonly
  """

  def __init__(self, tool, email, apikey=None, apikey_var=None, threads=None, qid=None):
    """:ivar result: :class:`entrezpy.base.result.EutilsResult`"""
    super().__init__('efetch.fcgi', tool, email, apikey, threads, qid=qid)

  def inquire(self, parameter, analyzer=entrezpy.efetch.efetch_analyzer.EfetchAnalyzer()):
    """Implements :meth:`entrezpy.base.query.EutilsQuery.inquire` and configures
    fetch.

    .. note :: Efetch prefers to know the number of UIDs to fetch, i.e. number
      of UIDs or retmax. If this information is missing, the max number
      of UIDs for the specific retmode and rettype are fetched.

    :param dict parameter: EFetch parameter
    :param analyzer: analyzer for Efetch results
    :type analyzer: :class:`entrezpy.base.analyzer.EutilsAnalyzer`
    :return: analyzer instance or None if request errors have been encountered
    :rtype: :class:`entrezpy.base.analyzer.EutilsAnalyzer` or None
    """
    param = entrezpy.efetch.efetch_parameter.EfetchParameter(parameter)
    logger.debug(json.dumps({__name__ : {'Parameter' : param.dump()}}))
    self.monitor_start(param)
    req_size = param.reqsize
    for i in range(param.expected_requests):
      if i * req_size + req_size > param.retmax:
        req_size = param.retmax % param.reqsize
      self.add_request(entrezpy.efetch.efetch_request.EfetchRequest(self.eutil,
                                                                    param,
                                                                    (i*param.reqsize),
                                                                    req_size), analyzer)
    self.request_pool.drain()
    self.monitor_stop()
    if self.check_requests() == 0:
      return analyzer
    return None


  def check_requests(self):
    """Test for request errors

      :return: 1 if request errors else 0
      :rtype: int
    """
    if not self.hasFailedRequests():
      logger.info(json.dumps({__name__ : {'Query status' : {self.id : 'OK'}}}))
      return 0
    logger.info(json.dumps({__name__ : {'Query status' : {self.id : 'failed'}}}))
    logger.debug(json.dumps({__name__ : {'Query status' :
                                         {self.id : 'failed',
                                          'request-dumps' : [x.dump_internals()
                                                             for x in self.failed_requests]}}}))
    return 1



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: bare
  :synopsis: Exports the base LinkSet class for Elinker() results.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""

from entrezpy.elink.linkset.unit import neighbor
from entrezpy.elink.linkset.unit import neighbor_score
from entrezpy.elink.linkset.unit import neighbor_history
from entrezpy.elink.linkset.unit import linklist
from entrezpy.elink.linkset.unit import linkin
from entrezpy.elink.linkset.unit import linkout
from entrezpy.elink.linkset.unit import linkout_nonlib_attribute
from entrezpy.elink.linkset.unit import linkout_all_attribute
from entrezpy.elink.linkset.unit import linkout_provider

class LinkSet:
  """The LinkSet class implements the base class for all Elink result outputs.

  ELink reports relations between UIDs from a source database (dbfrom) and a
  target database (dbto), e.g. which UID from the 'gene' database corresponds
  to a protein sequence in the 'protein' database.

  The LinkSet class stores and return references to classes implementing the
  result of specific Elink commands, e.g. 'cmd=neighbor'. These are stored in
  link_linksets and list_linksets to help the ElinkAnalyzer use the proper
  parser.

  LinkSet instances store the information about a linked database as a
  LinkSetUnit instance in its linkunits list. LinkSet instances can occur in
  two categories: linked and relaxed.

    Linked represent one-to-many UID relationships where one UID links to
    several UIDs. This occurs when UID in the db parameter are concatenated by
    id, i.e. id=19880848&id=19822630. Elinker tries to use this behavior by
    default.

    Relaxed represent many-to-many relationships where several source UID are
    linked to several target uids. This occurs when UIDs in the db parameter
    are concatenated by commas, i.e. id=19880848,19822630. This can be forced
    by setting the Elinker parameter 'linked' to False.

  :param str category: LinKSet category
  :param str dbfrom: name of database to link from
  :param bool canLink: linkunits can be used for automated follow-up parameter
  """

  link_linksets = {'neighbor' : neighbor.Neighbor,
                   'neighbor_score' : neighbor_score.NeighborScore,
                   'neighbor_history' : neighbor_history.NeighborHistory}

  list_linksets = {'acheck' : linklist.LinkList,
                   'ncheck' : linkin.LinkIn,
                   'lcheck' : linkout.LinkOut,
                   'llinks' : linkout_nonlib_attribute.LinkOutNonlibAttributes,
                   'llinkslib' : linkout_all_attribute.LinkOutAllAttribute,
                   'prlinks' : linkout_provider.LinkOutProvider}

  @staticmethod
  def new_unit(elink_cmd):
    """ Returns class reference implementing ELink command result unit.

    :param str elink_cmd: Elink command
    """
    if elink_cmd in LinkSet.link_linksets:
      return LinkSet.link_linksets[elink_cmd]
    if elink_cmd in LinkSet.list_linksets:
      return LinkSet.list_linksets[elink_cmd]
    return None

  def __init__(self, category, dbfrom, canLink):
    """The class LinkSet represent a Elink result. One Elink results corresponds
    to one Elink unit.

    :ivar list linkunits: storage for individual Elink results
    """
    self.category = category
    self.db = dbfrom
    self.linkunits = []
    self.canLink = canLink

  def add_linkunit(self, lsetunit):
    """:param lsetunit: populated Linkset unit
       :type  lsetunit: :class:`entrezpy.elink.linkset.unit.LinksetUnit
    """
    self.linkunits.append(lsetunit)

  def size(self):
    """:rtype: int"""
    return len(self.linkunits)

  def base_dump(self):
    """Dump common instance attributes

    :rtype: dict
    """
    return {'dbfrom' : self.db, 'category' : self.category,
            'size' : self.size(), 'linkunits' : [x.dump() for x in self.linkunits]}

  def dump(self):
    """ Virtual function to dump attributes in derived instances."""
    raise NotImplementedError()



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: linkout_provider
  :synopsis: Exports class LinkIn implementing Elink results from the Elink
    `prlinks` command.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import entrezpy.elink.linkset.unit.linksetunit


class LinkOutProvider(entrezpy.elink.linkset.unit.linksetunit.LinksetUnit):
  """The `LinkOutProvider` class represents a result from the Elink command
  `prlinks`

  :param dict unit: collected information for one linkout provider
  """
  @classmethod
  def new(cls, unit):
    """Returns new instance.

    :rtype: `entrezpy.elink.linkset.unit.linksetunit.linkin.LinkIn`
    """
    return cls(unit)

  @staticmethod
  def set_url(urlobj):
    """Converts an UrlObj url and  language info in two dicts into single dict

    :param dict urlobj: Elink url JSON result
    :rtype: dict
    """
    if not urlobj:
      return None
    url = urlobj.pop('value', None)
    if not url:
      return None
    return {url : urlobj.pop('lng', None)}


  class Provider:
    """Provider implements the provider information providded in the results

    :param dict provider_obj: provider information
    """
    def __init__(self, provider_obj):
      self.id = int(provider_obj.pop('id'))
      self.name = provider_obj.pop('name', None)
      self.nameabbr = provider_obj.pop('nameabbr', None)
      self.url = LinkOutProvider.set_url(provider_obj.pop('url', None))

    def dump(self):
      """:rtype: dict"""
      return {'id' : self.id, 'name' : self.name, 'nameabbr' : self.nameabbr, 'url' : self.url}

  def __init__(self, unit):
    super().__init__(None, unit.pop('linkname', None))
    self.url = unit['url'].pop('value', None)
    self.iconurl = LinkOutProvider.set_url(unit.pop('iconurl', None))
    self.subjecttypes = unit.pop('subjecttypes', None)
    self.categories = unit.pop('categories', None)
    self.attributes = unit.pop('attributes', None)
    self.provider = self.Provider(unit.pop('provider', None))

  def dump(self):
    """:return: dict"""
    return dict({'iconurl' : self.iconurl, 'provider' : self.provider.dump(),
                 'categories' : self.categories, 'attributes' : self.attributes,
                 'subjecttypes' : self.subjecttypes, }, **self.basic_dump())



"""
..
  Copyright 2018 The University of Sydney

  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.base.query
  :synopsis: Exports the base class for entrezpy queries to NCBI E-Utils

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import os
import uuid
import base64
import atexit
import queue
import threading

import entrezpy.requester.requester
import entrezpy.base.monitor


class EutilsQuery:
  """ EutilsQuery implements the base class for all entrezpy queries to
  E-Utils. It handles the information required by every query, e.g. base query
  url, email address, allowed requests per second, apikey,  etc. It declares
  the virtual method :meth:`.inquire` which needs to be implemented by every
  request since they differ among queries.

  An NCBI API key will bet set as follows:

  - passed as argument during initialization
  - check enviromental variable passed as argument
  - check enviromental variable NCBI_API_KEY

  Upon initalization, following parameters are set:

  - set unique query id
  - check for / set NCBI apikey
  - initialize :class:`entrezpy.requester.requester.Requester` with allowed
    requests per second
  - assemble Eutil url for desire EUtils function
  - initialize Multithreading queue and register query at
    :class:`entrezpy.base.monitor.QueryMonitor` for logging

  Multithreading is handled using the nested classes
  :class:`entrezpy.base.query.EutilsQuery.RequestPool` and
  :class:`entrezpy.base.query.EutilsQuery.ThreadedRequester`."""

  base_url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils"
  """Base url for all Eutil request"""

  query_requester = None
  """References :class:`entrezpy.requester.requester.Requester` """

  query_monitor = entrezpy.base.monitor.QueryMonitor()
  """References :class:`entrezpy.base.monitor.QueryMonitor` """

  @staticmethod
  def run_one_request(request):
    """Processes one request from the queue and logs its progress.

    :param request: single entrezpy request
    :type  request: :class:`entrezpy.base.request.EutilsRequest`
    """
    request.start_stopwatch()
    o = EutilsQuery.query_monitor.get_observer(request.query_id)
    o.observe(request)
    response = EutilsQuery.query_requester.request(request)
    request.calc_duration()
    o.processed_requests += 1
    return response

  class RequestPool:
    """ Threading Pool for requests. This class inits the threading pool,
    adds requests waits until all threads finish. A request consist of a tuple
    with the request and corresponding analyzer. Failed requests are stored
    separately to handle them later. If the number of threads is 0, use
    :meth:`entrezpy.base.query.EutilsQuery.RequestPool.run_single`. Otherwise,
    call :class:`entrezpy.base.query.EutilsQuery.ThreadedRequester`
    This is useful in cases where analyzers are calling not thread-safe methods
    or classes, e.g. Sqlite3
    """

    def __init__(self, num_threads, failed_requests):
      """Initiates a threading pool with a given number of threads.

      :param int num_threads: number of threads
      :param reference failed_requests:
        :attr:`entrezpy.base.query.EutilsQuery.failed_requests`
      :ivar requests: request queue
      :type requests: :class:`queue.Queue`
      :ivar bool useThreads: flag to use single ot multithreading
      """
      self.requests = queue.Queue(num_threads)
      self.failed_requests = failed_requests
      atexit.register(self.destructor)
      self.useThreads = True if num_threads > 0 else False
      if self.useThreads:
        for _ in range(num_threads):
          EutilsQuery.ThreadedRequester(self.requests, self.failed_requests)

    def add_request(self, request, analyzer):
      """Adds one request into the threading pool as
      **tuple**\ (`request`, `analzyer`).

      :param  request: entrezpy request instance
      :type   request: :class:`entrezpy.base.request.EutilsRequest`
      :param analyzer: entrezpy analyzer instance
      :type  analyzer: :class:`entrezpy.base.analyzer.EutilsAnalyzer`
      """
      self.requests.put((request, analyzer))

    def drain(self):
      """Empty threading pool and wait until all requests finish"""
      if self.useThreads:
        self.requests.join()
      else:
        self.run_single()

    def run_single(self):
      """Run single threaded requests."""
      while not self.requests.empty():
        request, analyzer = self.requests.get()
        response = EutilsQuery.run_one_request(request)
        if response:
          analyzer.parse(response, request)
        else:
          self.failed_requests.append(request)

    def destructor(self):
      """ Shutdown all ongoing threads when exiting due to an error.

      .. note::
        Deamon processes don't always stop when the main
        program exits and hang aroud. atexit.register(self.desctructor) seems
        to be a way to implement a dectructor. Currently not used.
      """
      pass

  class ThreadedRequester(threading.Thread):
    """ThreadedRequester handles multitthreaded request. It inherits from
    :class:`threading.Thread`. Requests are fetched  from
    :class:`entrezpy.base.query.EutilsQuery.RequestPool` and processed in
    :meth:`.run`.
    """
    def __init__(self, requests, failed_requests):
      """Inits :class:`.ThreadedRequester` to handle multithreaded requests.

      :param reference requests:
        :attr:`entrezpy.base.query.EutilsQuery.RequestPool.requests`
      :type reference failed_request:
        :attr:`entrezpy.base.query.EutilsQuery.failed_requests`
      """
      super().__init__(daemon=True)
      self.requests = requests
      self.failed_requests = failed_requests
      self.start()

    def run(self):
      """Overwrite :meth:`threading.Thread.run` for multithreaded requests."""
      while True:
        request, analyzer = self.requests.get()
        response = EutilsQuery.run_one_request(request)
        if response:
          analyzer.parse(response, request)
        else:
          self.failed_requests.append(request)
        self.requests.task_done()

  def __init__(self, eutil, tool, email, apikey=None, apikey_var=None, threads=None, qid=None):
    """Inits EutilsQuery instance with eutil, toolname, email, apikey,
    apikey_envar, threads and qid.

    :param str eutil: name of eutil function on EUtils server
    :param str tool: tool name
    :param str email: user email
    :param str apikey: NCBI apikey
    :param str apikey_var: enviroment variable storing NCBI apikey
    :param int threads: set threads for multithreading
    :param str qid: unique query id

    :ivar id: unique query id
    :ivar base_url: unique query id
    :ivar int requests_per_sec:  default limit of requests/sec (set by NCBI)
    :ivar int max_requests_per_sec:  max.requests/sec with apikeyby (set NCBI)
    :ivar str url:  full URL for Eutil function
    :ivar str contact:  user email (required by NCBI)
    :ivar str tool:  tool name (required by NCBI)
    :ivar str apikey:  NCBI apikey
    :ivar int num_threads:  number of threads to use
    :ivar list failed_requests: store failed requests for analysis if desired
    :ivar request_pool: :class:`entrezpy.base.query.EutilsQuery.RequestPool` instance
    :ivar int request_counter: requests counter for a EutilsQuery instance
    """
    self.id = base64.urlsafe_b64encode(uuid.uuid4().bytes).decode('utf-8') if not qid else qid
    self.eutil = eutil
    self.requests_per_sec = 3
    self.max_requests_per_sec = 10
    self.url = '/'.join([EutilsQuery.base_url, self.eutil])
    self.contact = email
    self.tool = tool
    self.apikey = self.check_ncbi_apikey(apikey, apikey_var)
    self.num_threads = 0 if not threads else threads
    self.failed_requests = []
    self.request_pool = EutilsQuery.RequestPool(self.num_threads, self.failed_requests)
    self.request_counter = 0
    EutilsQuery.query_requester = entrezpy.requester.requester.Requester(1/self.requests_per_sec)
    EutilsQuery.query_monitor.register_query(self)

  def inquire(self, parameter, analyzer):
    """Virtual function starting query. Each query requires its own implementation.

    :param dict parameter: E-Utilities parameters
    :param analzyer: query response analyzer
    :type  analzyer: :class:`entrezpy.base.analyzer.EutilsAnalzyer`
    :returns: analyzer
    :rtype: :class:`entrezpy.base.analyzer.EutilsAnalzyer`
    """
    raise NotImplementedError("{} requires inquire() implementation".format(__name__))

  def check_requests(self):
    """Virtual function testing and handling failed requests. These requests
    fail due to HTTP/URL issues and stored
    :attr:`entrezpy.base.query.EutilsQuery.failed_requests`
    """
    raise NotImplementedError("{} requires check_failed_requests() implementation".format(__name__))

  def check_ncbi_apikey(self, apikey=None, env_var=None):
    """Checks and sets NCBI apikey.

    :param str apikey: NCBI apikey
    :param str env_var: enviromental variable storing NCBI apikey
    """
    if 'NCBI_API_KEY' in os.environ:
      self.requests_per_sec = self.max_requests_per_sec
      return os.environ['NCBI_API_KEY']
    if apikey:
      self.requests_per_sec = self.max_requests_per_sec
      return apikey
    if env_var and (env_var in os.environ):
      self.requests_per_sec = self.max_requests_per_sec
      return os.environ[env_var]
    return None

  def prepare_request(self, request):
    """Prepares request for sending to E-Utilities with require quey attributes.

    :param request: entrezpy request instance
    :type  request: :class:`entrezpy.base.request.EutilsRequest`
    :returns: request instance with EUtils parameters
    :rtype: :class:`entrezpy.base.request.EutilsRequest`
    """
    request.id = self.request_counter
    request.query_id = self.id
    request.contact = self.contact
    request.url = self.url
    request.tool = self.tool
    request.apikey = self.apikey
    return request

  def add_request(self, request, analyzer):
    """Adds one request and corresponding analyzer to the request pool.

    :param request: entrezpy request instance
    :type  request: :class:`entrezpy.base.request.EutilsRequest`
    :param analzyer: entrezpy analyzer instance
    :type analyzer: :class:`entrezpy.base.analzyer.EutilsAnalyzer`
    """
    self.request_pool.add_request(self.prepare_request(request), analyzer)
    self.request_counter += 1

  def monitor_start(self, query_parameters):
    """Starts query monitoring

    :param query_parameters: query parameters
    :type query_parameters: :class:`entrezpy.base.parameter.EutilsParameter`
    """
    EutilsQuery.query_monitor.dispatch_observer(self, query_parameters)

  def monitor_stop(self):
    """Stops query monitoring"""
    EutilsQuery.query_monitor.recall_observer(self)

  def monitor_update(self, updated_query_parameters):
    """Updates query monitoring parameters if follow up requests are required.

    :param updated_query_parameters: updated query parameters
    :type  updated_query_parameters: :class:`entrezpy.base.parameter.EutilsParameter`
    """
    EutilsQuery.query_monitor.update_observer(self, updated_query_parameters)

  def hasFailedRequests(self):
    """Reports if at least one request failed."""
    if self.failed_requests:
      return True
    return False

  def dump(self):
    """Dump all attributes"""
    return {'id' : self.id, 'base_url' : EutilsQuery.base_url, 'eutil' : self.eutil,
            'url' : self.url, 'req/sec' : self.requests_per_sec, 'tool' : self.tool,
            'contact' : self.contact, 'apikey' : self.apikey, 'threads' : self.num_threads}



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.elink.linkset.unit.linkout_nonlib_attributes
   :synopsis: Exports :class:`LinkOutNonlibAttributes` implementing Elink
    results for `llinks` command reporting URLs and attributes for all links
    outside Entrez not being libraries.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import entrezpy.elink.linkset.unit.linksetunit


class LinkOutNonlibAttributes(entrezpy.elink.linkset.unit.linksetunit.LinksetUnit):
  """ The `LinkOutNonlibAttribute` class represents a result from the `llinks`
  Elink command. Results show the URL and its attributes for all links outside
  Entrez which are not libraries.

  :param dict objurl: Entrez link data
  """
  @classmethod
  def new(cls, objurl):
    """:rtype: :class:`linkout_nonlib_attributes.LinkOutNonlibAttributes`"""
    return cls(objurl)

  def __init__(self, objurl):
    super().__init__(None, None)
    self.objurl = objurl

  def dump(self):
    ":rtype: dict"
    return dict({'objurl' : self.objurl}, **self.basic_dump())



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.esearch.esearch_result
  :synopsis: Exports class EsearchResult implementing entrezpy results from
    NCBI Esearch Eutils requests


.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import entrezpy.base.result


class EsearchResult(entrezpy.base.result.EutilsResult):
  """ EsearchResult sstores fetched UIDs and/or WebEnv-QueryKeys and
  creates follow-up parameters. UIDs are stored as string, even when UIDs,
  since responses can contain also accsessions when using the idtype option.

  :param dict response: Esearch response
  :param request: Esearch request instance for this query
  :type request: :class:`entrezpy.esearch.esearch_request.EsearchRequest`
  :ivar list uids: analyzed UIDs from response
  """

  def __init__(self, response, request):
    super().__init__('esearch', request.query_id, request.db, response.get('webenv'),
                     response.pop('querykey', None))
    self.count = int(response.get('count'))
    self.retmax = int(response.pop('retmax'))
    self.retstart = int(response.pop('retstart'))
    self.uids = response.pop('idlist', [])

  def dump(self):
    """:rtype: dict"""
    return {'db':self.db, 'count' : self.count, 'len_uids' : len(self.uids),
            'uid' : self.uids, 'retmax' : self.retmax, 'function':self.function,
            'retstart' : self.retstart, 'references' : self.references.dump()}

  def get_link_parameter(self, reqnum=0):
    """Assemble follow-up parameters for linking. The first request returns
    all required information and using its querykey in such a case.

    :rtype: dict
    """
    if self.uids:
      return {'db' : self.db, 'id' : self.uids, 'WebEnv' : self.webenv,
              'query_key' : self.references.get_querykey(self.webenv, reqnum)}
    retmax = self.retmax
    if retmax == 0:
      retmax = self.count
    return {'db' : self.db, 'WebEnv' : self.webenv, 'retmax' : retmax,
            'query_key' : self.references.get_querykey(self.webenv, reqnum)}

  def isEmpty(self):
    """Empty search result has no webenv/querykey and/or no fetched UIDs"""
    if self.references.size() > 0:
      return False
    if self.uids:
      return False
    return True

  def size(self):
    """Get number of analyzed UIDs

    :rtype: int
    """
    if self.uids:
      return len(self.uids)
    return self.count


  def query_size(self):
    """Get number of all UIDs for search (count)

    :rtype: int
    """
    return self.count

  def add_response(self, response):
    """Add responses from individual requests

    :param dict response: Esearch response
    """
    self.references.add_reference(response.pop('webenv', None), response.pop('querykey', None))
    self.uids += response.pop('idlist', [])



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: elinker
   :synopsis: Exports ELinker class implementing Elink queries.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json
import logging

import entrezpy.base.query
import entrezpy.elink.elink_parameter
import entrezpy.elink.elink_request
import entrezpy.elink.elink_analyzer

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())

class Elinker(entrezpy.base.query.EutilsQuery):
  """Elinker implements elink queries to E-Utilities [0].
  Elinker implements the inquire() method to link data sets on NCBI Entrez
  servers. All parameters described in [0] are acccepted. Elink queries consist
  of one request linking UIDs or an earlier requests on the history server
  within the same or different Entrez database. [0]:
  https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.ELink

  :param str tool: tool name
  :param str email: user email
  :param str apikey: NCBI apikey
  :param str apikey_var: enviroment variable storing NCBI apikey
  :param int threads: set threads for multithreading
  :param str qid: unique query id
  """

  def __init__(self, tool, email, apikey=None, apikey_var=None, threads=None, qid=None):
    super().__init__('elink.fcgi', tool, email, apikey, apikey_var, threads, qid)

  def inquire(self, parameter, analyzer=entrezpy.elink.elink_analyzer.ElinkAnalyzer()):
    """ Implements virtual function inquire()

      1. Prepares parameter instance :class:`entrezpy.elink.elink_parameter.ElinkerParameter`
      2. Starts threading monitor :func:`monitor_start`
      3. Adds ElinkRequests to queue :func:`add_request`
      4. Runs and analyzes all requests
      5. Checks for errors :func:`check_requests`

    :param dict parameter: ELink parameter
    :param analyzer analyzer: analyzer for Elink Results, default is
      :class:`entrezpy.elink.elink_analyzer.ElinkAnalyzer`
    :return: analyzer  or None if request errors have been encountered
    :rtype: :class:`entrezpy.base.analyzer.EntrezpyAnalyzer` instance or None
    """
    logger.debug(json.dumps({__name__ : {'dump' : self.dump()}}))
    p = entrezpy.elink.elink_parameter.ElinkParameter(parameter)
    self.monitor_start(p)
    self.add_request(entrezpy.elink.elink_request.ElinkRequest(self.eutil, p), analyzer)
    self.request_pool.drain()
    self.monitor_stop()
    if self.check_requests() == 0:
      return analyzer
    return None

  def check_requests(self):
    """Test for request errors

      :return: 1 if request errors else 0
      :rtype: int
    """
    if not self.hasFailedRequests():
      logger.info(json.dumps({__name__ : {'Query status' : {self.id : 'OK'}}}))
      return 0
    logger.info(json.dumps({__name__ : {'Query status' : {self.id : 'failed'}}}))
    logger.debug(json.dumps({__name__ : {'Query status' :
                                         {self.id : 'failed',
                                          'request-dumps' : [x.dump_internals()
                                                             for x in self.failed_requests]}}}))
    return 1



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.esearch.esearch_parameter
  :synopsis: Export EsearchParameter for entrezpy queries to NCBI E-Utils
    Esearch

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import sys
import math
import json
import logging

import entrezpy.base.parameter


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())


MAX_REQUEST_SIZE = 100000
"""Maximum number of UIDs for one request"""

class EsearchParameter(entrezpy.base.parameter.EutilsParameter):
  """EsearchParameter checks query specific parameters and configures an Esearch
  query. If more than one request is required the instance is reconfigured by
  :meth:`entrezpy.esearch.esearcher.Esearcher.configure_follow_up`.

  .. note :: EsearchParameter works best when using the NCBI Entrez history
    server. If usehistory is not used, linking requests cannot be guaranteed.
  """

  def __init__(self, parameter):
    super().__init__(parameter)
    self.retmode = 'json'
    self.uilist = self.set_uilist(parameter.get('rettype'))
    self.retmax = self.adjust_retmax(parameter.get('retmax'))
    self.reqsize = self.adjust_reqsize(int(parameter.get('reqsize', MAX_REQUEST_SIZE)))
    self.retstart = int(parameter.get('retstart', 0))
    self.usehistory = parameter.get('usehistory', True)
    self.term = parameter.get('term')
    self.sort = parameter.get('sort')
    self.field = parameter.get('field')
    self.datetype = parameter.get('datetype')
    self.reldate = parameter.get('reldate')
    self.mindate = parameter.get('mindate')
    self.maxdate = parameter.get('maxdate')
    self.idtype = parameter.get('idtype')
    self.check()

  def goodDateparam(self):
    """:rtype: bool"""
    useDate = False
    if self.useMinMaxDate():
      if self.reldate:
        logger.error(json.dumps({__name__: {'Error' : 'Cannot use reldate and Min/Max dates'}}))
        return False
      useDate = True
    if self.reldate:
      useDate = True
    if useDate and (not self.datetype):
      logger.error(json.dumps({__name__: {'Error' : 'Require datetype with dates'}}))
      return False
    return True

  def useMinMaxDate(self):
    """:rtype: bool"""
    if self.mindate or self.maxdate: # Intend to use max/min dates
      if self.mindate and self.maxdate: # Require both
        return True
      logger.error(json.dumps({__name__: {'Error' : 'Require mindate and maxdate'}}))
    return False

  def set_uilist(self, rettype):
    """:rtype: bool"""
    if not rettype or (rettype == 'uilist'):
      return True
    return False

  def adjust_retmax(self, retmax):
    """Adjusts retmax parameter. Order of check is crucial.

    :param int retmax: retmax value
    :return: adjusted retmax
    :rtype: int
    """
    if not self.uilist:
      return 0
    if retmax is None:
      return None
    if int(retmax) == 0:
      return 0
    return int(retmax)

  def adjust_reqsize(self, request_size):
    """Adjusts request size for low retmax

    :return: adjusted request size
    :rtype: int
    """
    if self.retmax is None:
      return request_size
    if self.retmax < request_size:
      return self.retmax
    return request_size

  def calculate_expected_requests(self, qsize=None, reqsize=None):
    """Calculate anf set the expected number of requests. Uses internal
    parameters if non are provided.

    :param int or None qsize: query size, i.e. expected number of data sets
    :param int reqsize: number of data sets  to fetch in one request
    """
    if not qsize:
      qsize = self.retmax
    if not reqsize:
      reqsize = MAX_REQUEST_SIZE
    self.expected_requests = math.ceil(qsize / reqsize)

  def check(self):
    """Implements :class:`entrezpy.base.parameter.EutilsParameter.check` to
    check for the minumum required parameters. Aborts if any check fails.
    """
    if not self.haveDb():
      sys.exit(logger.error(json.dumps({__name__: {'Error' : {'Missing parameter': 'db'},
                                                   'action' : 'abort'}})))

    if not self.haveExpectedRequets():
      sys.exit(logger.error(json.dumps({__name__ : {'Error' :
                                                    {'expected requests' : self.expected_requests},
                                                    'action' : 'abort'}})))

    if not self.goodDateparam():
      sys.exit(logger.error(json.dumps({__name__ : {'Error' : 'Bad date parameters',
                                                    'action' : 'abort'}})))

    if not self.term and not self.webenv:
      sys.exit(logger.error(json.dumps({__name__:{'Error': {'Missing  parameters' : 'term, WebEnv'},
                                                  'action' : 'abort'}})))

  def dump(self):
    return {'db' : self.db, 'webenv' : self.webenv, 'querykey' : self.querykey,
            'usehistory' : self.usehistory, 'term':self.term, 'retmode':self.retmode,
            'uilist': self.uilist, 'retmax': self.retmax, 'retstart': self.retstart,
            'sort' : self.sort, 'field' : self.field, 'datetype' : self.datetype,
            'reldate' : self.reldate, 'mindate' : self.mindate, 'maxdate' : self.maxdate,
            'expected_requsts' : self.expected_requests,
            'request_size' : self.reqsize, 'max_request_size' : MAX_REQUEST_SIZE}



# Copyright 2018, 2019 The University of Sydney
# This file is part of entrezpy.
#
#  Entrezpy is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Lesser General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.
"""
.. module:: logging
  :synopsis:
    This module is part of entrezpy. It configures logging via Python's :mod:`logging` implements the base class for all
    entrezpy analzyer for NCBI E-Utils responses.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""

import os
import time
import logging

default_config = {
  'disable_existing_loggers': False,
  'version': 1,
  'formatters': {
    'short': { 'format': '%(asctime)s %(levelname)s %(name)s: %(message)s'},
  },
  'handlers': {
    'console': {
      'level': 'INFO',
      'formatter': 'short',
      'class': 'logging.StreamHandler',
    },
    'file': {
      'level': 'DEBUG',
      'formatter': 'short',
      'class': 'logging.FileHandler',
      'filename' : os.path.join(os.getcwd(), 'entrepy-{}.log'.format(time.time())),
      'delay' : True
    },
  },
  'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'ERROR',
        },
    'plugins': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': False
        }
    },
}



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.base.result
  :synopsis: Exports the base class for entrezpy results from NCBI E-Utils
    queries

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""

import entrezpy.base.referencer


class EutilsResult:
  """EutilsResult is the base class for an entrezpy result. It sets the
  required result attributes common for all result and declares virtual
  functions to interact with other entrezpy classes. Empty results are
  successful results since no query error has been received.
  :meth:`entrezpy.base.result.EutilsResult.size` is important to

  - determine if and how many follow-up requests are required
  - if it's an empty result

  :param string function: EUtil function of the result
  :param string qid: query id
  :param string db: Entrez database name for result
  :param string webenv: WebEnv of response
  :param int querykey: querykey of response
  """

  def __init__(self, function, qid, db, webenv=None, querykey=None):
    self.function = function
    self.query_id = qid
    self.db = db
    self.webenv = webenv
    self.references = entrezpy.base.referencer.EutilReferencer(self.webenv, querykey)

  def size(self):
    """Returns result size in the corresponfong ResultSize unit

    :rtype: int
    :raises NotImplementedError: if implementation is missing"""
    raise NotImplementedError("Help! Require implementation")

  def dump(self):
    """Dumps all instance attributes

    :rtype: dict
    :raises NotImplementedError: if implementation is missing"""
    raise NotImplementedError("Help! Require implementation")

  def get_link_parameter(self, reqnum=0):
    """Assembles parameters for automated follow-ups. Use the query key from
    the first request by default.

    :param int reqnum: request number for which query_key should be returned
    :return: EUtils parameters
    :rtype: dict
    :raises NotImplementedError: if implementation is missing"""
    raise NotImplementedError("Help! Require implementation")

  def isEmpty(self):
    """Indicates empty result.

    :rtype: bool
    :raises NotImplementedError: if implementation is missing"""
    raise NotImplementedError("Help! Require implementation")



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.elink.linkset.relaxed
   :synopsis:
    Exports RelaxedLinkset class implementing many-to-many Elink results.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""

import entrezpy.elink.linkset.bare

class RelaxedLinkset(entrezpy.elink.linkset.bare.LinkSet):
  """
  The RelaxedLinkset class represents a collection of Elink results where
  several UID's from a source database (dbfrom) are linked to several UID's
  from a target database (fromdb). This usually occurs when creating elink
  commands with one id parameter concatenating several UIDs by a comma, i.e.
  id=19880848,19822630.

  :param uidsfrom: UIDs from database to link from
  :type uidsfrom: list
  :param dbfrom: name of database to link from
  :type dbfrom: str
  :param canLink: linkunits can be used for automated follow-up parameter
  :type canLink: boolean
  """

  def __init__(self, uidsfrom, dbfrom, canLink=True):
    """ :attr dict uids: UIDs from source database"""
    super().__init__('relaxed', dbfrom, canLink)
    self.uids = {int(x) : 0 for x in uidsfrom}


  def get_link_uids(self):
    """:return: target database and its linl UIDs
       :rtype: dict
    """
    link = {}
    for i in self.linkunits:
      if i.db not in link:
        link[i.db] = []
      link[i.db].append(i.uid)
    return link
    #return [x.uids for x in self.linkunits]

  def dump(self):
    """:return: all basis attributes of the instance
    :rtype: dict
    """
    return dict({'src_uids': [x for x in self.uids]},**self.base_dump())



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: linksetunit
  :synopsis: Exports class LinksetUnit implementing result units for LinkSet().

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


class LinksetUnit:
  """
  The LinksetUnit class implements the base class for all LinksetUnits.
  LinksetUnits atore the information for UIDs linked to one or several source
  UIDs. LinksetUnit instances are handled by LinkSet instances. Almost all
  LinksetUnits have a dbto and linkname parameter. Some exceptions exists and
  these parameters are then set to None.

  :param str dbto: name of linked database
  :param str linkname: linkname
  """

  def __init__(self, dbto, linkname, cat='basic'):
    """Inits LinksetUnit instance with the linked database name and linkname

    :attribute str dbto: name of target database
    :attribute str linkname: ELink linkname
    """
    self.db = dbto
    self.linkname = linkname
    self.cat = cat

  def dump(self):
    """Virtual function to dump attributes in derived instances.

      :return: all LinksetUnit instance attributes
      :rtype: dict
    """
    raise NotImplementedError()

  def basic_dump(self):
    """:return: basis attributes of LinksetUnit instance
       :rtype: dict
    """
    return {'dbto' : self.db, 'linkname' : self.linkname}
